from typing import TypeVar, Generic, Optional, Callable, Tuple, Any, cast

E = TypeVar('E')
T = TypeVar('T')
U = TypeVar('U')
R = TypeVar('R')
_T = TypeVar('_T')


class Option(Generic[T]):

    def __init__(self, val: Optional[T]):
        self.__val: Optional[T] = val

    @staticmethod
    def Some(val: _T) -> "Option[_T]":
        return Option(val=val)

    @staticmethod
    def None_() -> "Option[Any]":
        return Option(val=None)

    def is_some(self) -> bool:
        return self.__val is not None

    def is_some_and(self, f: Callable[[T], bool]) -> bool:
        return f(cast(T, self.__val)) if self.is_some() else False

    def is_none(self) -> bool:
        return not self.is_some()

    def expect(self, msg: str) -> T:
        if self.is_some():
            return cast(T, self.__val)
        else:
            raise ValueError(msg)

    def unwrap(self) -> T:
        if self.is_some():
            return cast(T, self.__val)
        else:
            raise ValueError("Option is None.")

    def unwrap_or(self, default: T) -> T:
        if self.is_some():
            return cast(T, self.__val)
        else:
            return default

    def unwrap_or_else(self, f: Callable[[], T]) -> T:
        if self.is_some():
            return cast(T, self.__val)
        else:
            return f()

    def map(self, f: Callable[[T], U]) -> 'Option[U]':
        if self.is_some():
            return Option.Some(f(cast(T, self.__val)))
        else:
            return cast(Option[U], Option.None_())

    def inspect(self, f: Callable[[T], None]):
        if self.is_some():
            f(cast(T, self.__val))
        return self

    def map_or(self, default: U, f: Callable[[T], U]) -> U:
        if self.is_some():
            return f(cast(T, self.__val))
        else:
            return default

    def map_or_else(self, default: Callable[[], U], f: Callable[[T], U]) -> U:
        if self.is_some():
            return f(cast(T, self.__val))
        else:
            return default()

    def ok_or(self, err: E):  # type: ignore
        from .rust_result import Result
        if self.is_some():
            return Result[T, E](val=cast(T, self.__val))
        else:
            return Result[T, E](err=err)

    def ok_or_else(self, err: Callable[[], E]):
        from .rust_result import Result
        if self.is_some():
            return Result[T, E](val=cast(T, self.__val))
        else:
            return Result[T, E](err=err())

    def and_(self, optb: 'Option[U]') -> 'Option[U]':
        if self.is_some():
            return optb
        else:
            return cast(Option[U], Option.None_())

    def and_then(self, f: Callable[[T], 'Option[U]']) -> 'Option[U]':
        if self.is_some():
            return f(cast(T, self.__val))
        else:
            return cast(Option[U], Option.None_())

    def filter(self, predicate: Callable[[T], bool]) -> 'Option[T]':
        if self.is_some():
            if predicate(cast(T, self.__val)):
                return Option.Some(cast(T, self.__val))
        return cast(Option[T], Option.None_())

    def or_(self, optb: 'Option[T]') -> 'Option[T]':
        if self.is_none():
            return optb
        return Option.Some(cast(T, self.__val))

    def or_else(self, f: Callable[[], 'Option[T]']) -> 'Option[T]':
        if self.is_none():
            return f()
        return Option.Some(cast(T, self.__val))

    def xor(self, optb: 'Option[T]') -> 'Option[T]':
        if self.is_some() and optb.is_none():
            return Option.Some(cast(T, self.__val))
        elif self.is_none() and optb.is_some():
            return optb
        else:
            return cast(Option[T], Option.None_())

    def zip(self, other: 'Option[U]') -> 'Option[Tuple[T, U]]':
        if self.is_some() and other.is_some():
            return Option.Some((cast(T, self.__val), cast(U, other.__val)))
        return cast(Option[Tuple[T, U]], Option.None_())

    def zip_with(self, other: 'Option[U]', f: Callable[[T, U], R]) -> 'Option[R]':
        if self.is_some() and other.is_some():
            return Option.Some(f(cast(T, self.__val), cast(U, other.__val)))
        return cast(Option[R], Option.None_())

    def unzip(self) -> 'Tuple[Option[Any], Option[Any]]':
        if self.is_some():
            if isinstance(self.__val, tuple):
                return Option.Some(self.__val[0]), Option.Some(self.__val[1])  # type: ignore
        return Option.None_(), Option.None_()



from typing import Any, TypeVar, Generic, Optional, Callable, cast

E = TypeVar("E")
T = TypeVar("T")
U = TypeVar("U")
_E = TypeVar("_E")
_T = TypeVar("_T")



class Result(Generic[T, E]):

    def __init__(self, val: Optional[T] = None, err: Optional[E] = None):
        self.__val = val
        self.__err = err

    @staticmethod
    def Ok(val: _T) -> "Result[_T, Any]":
        return Result(val=val)

    @staticmethod
    def Err(err: _E) -> 'Result[Any, _E]':
        return Result(err=err)

    def is_ok(self) -> bool:
        return self.__err is None

    def is_ok_and(self, f: Callable[[T], bool]) -> bool:
        if self.is_ok():
            return f(cast(T, self.__val))
        else:
            return False

    def is_err(self) -> bool:
        return not self.is_ok()

    def is_err_and(self, f: Callable[[E], bool]) -> bool:
        if self.is_ok():
            return False
        return f(cast(E, self.__err))

    def ok(self):
        from .rust_option import Option
        if self.is_ok():
            return Option[T](val=cast(T, self.__val))
        else:
            return Option[T](val=None)

    def err(self):
        from .rust_option import Option
        if self.is_err():
            return Option[E](val=cast(E, self.__err))
        else:
            return Option[E](val=None)

    def map(self, op: Callable[[T], U]) -> 'Result[U, E]':
        if self.is_ok():
            return Result[U, E](val=op(cast(T, self.__val)))
        else:
            return Result[U, E](err=cast(E, self.__err))

    def map_or(self, default: U, f: Callable[[T], U]) -> U:
        if self.is_ok():
            return f(cast(T, self.__val))
        else:
            return default

    def map_or_else(self, default: Callable[[E], U], f: Callable[[T], U]) -> U:
        if self.is_ok():
            return f(cast(T, self.__val))
        else:
            return default(cast(E, self.__err))

    def map_err(self, op: Callable[[E], U]) -> 'Result[T, U]':
        if self.is_ok():
            return Result[T, U](val=cast(T, self.__val))
        else:
            return Result[T, U](err=op(cast(E, self.__err)))

    def inspect(self, f: Callable[[T], None]) -> 'Result[T, E]':
        if self.is_ok():
            f(cast(T, self.__val))
        return self

    def inspect_err(self, f: Callable[[E], None]) -> 'Result[T, E]':
        if self.is_err():
            f(cast(E, self.__err))
        return self

    def expect(self, msg: str) -> T:
        if self.is_ok():
            return cast(T, self.__val)
        else:
            raise ValueError(msg)

    def expect_err(self, msg: str) -> E:
        if self.is_err():
            return cast(E, self.__err)
        else:
            raise ValueError(msg)

    def unwrap(self) -> T:
        if self.is_ok():
            return cast(T, self.__val)
        else:
            raise ValueError(f"Result is Err: {str(self.__err)}")

    def unwrap_or(self, default: T) -> T:
        if self.is_ok():
            return cast(T, self.__val)
        else:
            return default

    def unwrap_or_else(self, op: Callable[[E], T]) -> T:
        if self.is_ok():
            return cast(T, self.__val)
        else:
            return op(cast(E, self.__err))

    def unwrap_err(self) -> E:
        if self.is_err():
            return cast(E, self.__err)
        else:
            raise ValueError(f"Result is Ok: {str(self.__val)}")

    def and_(self, res: 'Result[U, E]') -> 'Result[U, E]':
        if self.is_ok():
            return res
        else:
            return Result[U, E](err=cast(E, self.__err))

    def and_then(self, op: Callable[[T], 'Result[U, E]']) -> 'Result[U, E]':
        if self.is_ok():
            return op(cast(T, self.__val))
        else:
            return Result[U, E](err=cast(E, self.__err))

    def or_(self, res: 'Result[T, E]') -> 'Result[T, E]':
        if self.is_ok():
            return Result[T, E](val=cast(T, self.__val))
        else:
            return res

    def or_else(self, op: Callable[[E], 'Result[T, E]']) -> 'Result[T, E]':
        if self.is_ok():
            return Result[T, E](val=cast(T, self.__val))
        else:
            return op(cast(E, self.__err))



from setuptools import setup

setup(
    name='pyruster',
    version='0.1.12',
    author='Elling',
    description='Implementing some syntax like rust.',
    long_description=open('README.rst').read(),
    license='MIT',
    install_requires=[
    ],
    packages=["pyruster", ],
    package_dir={"": "src"},
    package_data={"pyruster": ["py.typed", "*.pyi"]},
    python_requires=">=3.6",
)
